# -*- coding=utf-8 -*-

"""Unit testing for assembly graph."""

# pylint: disable=compare-to-zero, missing-raises-doc

from pathlib import Path

from pulp import LpStatus, LpStatusInfeasible

from khloraascaf.exceptions import (
    CombineScaffoldingError,
    NoSolution,
    NotACircuit,
    RepeatScaffoldingError,
    ScaffoldingError,
    SingleCopyScaffoldingError,
    UnfeasibleDR,
    UnfeasibleIR,
    UnfeasibleSC,
    WrongRegionID,
    WrongSolverName,
)
from khloraascaf.lib import IR_REGION_ID, SC_REGION_ID


# ============================================================================ #
#                                TEST FUNCTIONS                                #
# ============================================================================ #
def test_scaffolding_error():
    """Test ScaffoldingError exception."""
    exc = ScaffoldingError(Path('./jaaj'))
    assert str(exc) == (
        'Scaffolding fails\n'
        '\tSee output directory: jaaj'
    )


def test_repeat_scaffolding_error():
    """Test RepeatScaffoldingError exception."""
    exc = RepeatScaffoldingError()
    assert str(exc) == 'Repeat scaffolding has failed'


def test_singlecopy_scaffolding_error():
    """Test SingleCopyScaffoldingError exception."""
    exc = SingleCopyScaffoldingError()
    assert str(exc) == 'Single copy scaffolding has failed'


def test_combine_scaffolding_error():
    """Test CombineScaffoldingError exception."""
    exc = CombineScaffoldingError()
    assert str(exc) == 'The scaffolding combination has failed'


def test_wrong_region_code():
    """Test WrongRegionID exception."""
    exc = WrongRegionID(IR_REGION_ID)
    assert str(exc) == "The region identifier 'ir' is not correct"


def test_wrong_solver_name():
    """Test WrongSolverName exception."""
    exc = WrongSolverName('wrong_solver')
    assert str(exc) == 'The solver name wrong_solver is not correct'


def test_unfeasible_ir():
    """Test UnfeasibleIR exception."""
    exc = UnfeasibleIR(
        LpStatus[LpStatusInfeasible], (IR_REGION_ID, SC_REGION_ID),
    )
    assert str(exc) == (
        'The Find the best inverted repeats problem is unfeasible:\n'
        '\t* ILP codes: ir-sc\n'
        '\t* Status: Infeasible'
    )


def test_unfeasible_dr():
    """Test UnfeasibleDR exception."""
    exc = UnfeasibleDR(
        LpStatus[LpStatusInfeasible], (IR_REGION_ID, SC_REGION_ID),
    )
    assert str(exc) == (
        'The Find the best direct repeats problem is unfeasible:\n'
        '\t* ILP codes: ir-sc\n'
        '\t* Status: Infeasible'
    )


def test_unfeasible_sc():
    """Test UnfeasibleSC exception."""
    exc = UnfeasibleSC(
        LpStatus[LpStatusInfeasible], (IR_REGION_ID, SC_REGION_ID),
    )
    assert str(exc) == (
        'The Find the best single copy regions problem is unfeasible:\n'
        '\t* ILP codes: ir-sc\n'
        '\t* Status: Infeasible'
    )


def test_not_a_circuit():
    """Test NotACircuit exception."""
    exc = NotACircuit()
    assert str(exc) == 'The found path is not a circuit'


def test_no_solution():
    """Test NoSolution exception."""
    exc = NoSolution()
    assert str(exc) == 'There is no solution'



# -*- coding=utf-8 -*-

"""Module for assembly graph and multiple solution generator."""

from __future__ import annotations

from collections.abc import Iterable, Iterator
from pathlib import Path
from queue import LifoQueue

from bitarray import bitarray
from revsymg.graphs import RevSymGraph
from revsymg.index_lib import (
    FORWARD_INT,
    IND,
    ORIENT_REV,
    IndexT,
    IndOrIndT,
    IndOrT,
    OrT,
)

from khloraascaf.inputs import STR_ORIENT
from khloraascaf.lib import RegionIDT
from khloraascaf.outputs import (
    ORC_ID_IND,
    ORC_OR_IND,
    ORIENT_INT_STR,
    OrCT,
    read_contigs_of_regions,
    read_map_of_regions,
)
from khloraascaf.run_metadata import MetadataSolution


# DOCU: all assembly graph
# DOCU: tuto for assembly graph
# ============================================================================ #
#                                   CONSTANTS                                  #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                     Files                                    #
# ---------------------------------------------------------------------------- #
#
# Region paths
#
REGION_PATHS_PREFIX = 'region_paths'
"""Prefix of the region paths file name."""
REGION_PATHS_EXT = 'tsv'
"""Extension of the region paths file"""
#
# Oriented contig paths
#
ORIENTED_CONTIG_PATHS_PREFIX = 'oriented_contig_paths'
"""Prefix of the oriented contig paths file name."""
ORIENTED_CONTIG_PATHS_EXT = 'tsv'
"""Extension of the oriented contig paths file"""


# ============================================================================ #
#                                     CLASS                                    #
# ============================================================================ #
class AssemblyGraph():
    """Khloraascaf assembly graph."""

    @classmethod
    def from_solution_metadata(
            cls, solution_metadata: MetadataSolution) -> AssemblyGraph:
        """Instantiate an AssemblyGraph object from a solution metadata.

        Parameters
        ----------
        solution_metadata : MetadataSolution
            Solution metadata

        Returns
        -------
        AssemblyGraph
            Assembly graph associated to the solution
        """
        outdir = solution_metadata.solution_directory()
        return cls(
            outdir / solution_metadata.map_of_regions(),
            outdir / solution_metadata.contigs_of_regions(),
        )

    def __init__(self, map_of_regions_path: Path,
                 contigs_of_regions_path: Path):
        """The Initializer."""
        self.__graph: RevSymGraph = RevSymGraph()
        self.__regions_contigs: list[tuple[OrCT, ...]] = []
        self.__region_path_length: int = 0
        self.__read_regions_contigs(contigs_of_regions_path)
        self.__add_regions_links(map_of_regions_path)

    def all_region_paths(self) -> Iterator[tuple[IndOrT, ...]]:  # noqa
        """Iterate over all the paths of oriented regions.

        Yields
        ------
        tuple of IndOrT
            Path of oriented regions
        """
        edges = self.__graph.edges()

        used_edges: bitarray = bitarray('0') * (
            edges.biggest_edge_index() + 1)

        starter = (0, FORWARD_INT)

        path: list[IndOrT] = [starter]
        eind_path: list[IndexT] = []
        #
        # The lifo contains the number of elements to keep to continue
        #   at the next remaning branch (first IndexT value)
        #
        lifo: LifoQueue[tuple[IndexT, IndOrT, IndexT]] = LifoQueue()

        for v, e_ind in edges.succs(starter):
            if v[IND] == starter[IND]:
                yield tuple(path)
            else:
                lifo.put((1, v, e_ind))

        while not lifo.empty():
            _, v, e_ind = lifo.get()
            path.append(v)
            eind_path.append(e_ind)
            used_edges[e_ind] = True

            vw_not_used: list[IndOrIndT] = []
            for w, f_ind in edges.succs(v):
                if w[IND] != starter[IND] and not used_edges[f_ind]:
                    vw_not_used.append((w, f_ind))

            if vw_not_used:
                for w, f_ind in vw_not_used:
                    lifo.put((len(path), w, f_ind))
            else:
                if len(path) == self.__region_path_length:
                    yield tuple(path)
                #
                # Back to the next branching vertex
                #
                if not lifo.empty():
                    branch_src_succ_eind = lifo.get()
                    lifo.put(branch_src_succ_eind)
                    for _ in range(len(path) - branch_src_succ_eind[0]):
                        path.pop()
                        used_edges[eind_path.pop()] = False

    def all_oriented_contig_paths(self) -> Iterator[tuple[OrCT, ...]]:
        """Iterate over all the paths of oriented contigs.

        Yields
        ------
        tuple of IndOrT
            Path of oriented contigs
        """
        # TODO don't use self method, to avoid verify path validity
        # (because it is known valid)
        for region_path in self.all_region_paths():
            yield tuple(self.region_path_to_oriented_contigs(region_path))

    def region_path_to_oriented_contigs(
            self, region_path: Iterable[IndOrT]) -> Iterator[OrCT]:
        """Iterate over the oriented contigs of a given region path.

        Parameters
        ----------
        region_path : iterable of IndOrT
            Path of oriented regions

        Yields
        ------
        OrCT
            Oriented contig
        """
        # TODO raise error if not a valid path
        for reg_ind, reg_or in region_path:
            yield from self.oriented_contigs_of_region(reg_ind, reg_or)

    # ~*~ Getter ~*~

    def revsymg(self) -> RevSymGraph:
        """Return the reverse symmetric graph associated.

        Returns
        -------
        RevSymGraph
            Reverse symmetric graph
        """
        return self.__graph

    def oriented_contigs_of_region(self, region_ind: IndexT,
                                   orientation: OrT = FORWARD_INT) -> (
            Iterator[OrCT]):
        """Iterate over oriented contigs of the oriented region.

        Parameters
        ----------
        region_ind : IndexT
            Region's index
        orientation : OrT, optional
            Region's orientation, by default FORWARD_INT

        Yields
        ------
        OrCT
            Oriented contig of the oriented region
        """
        if orientation == FORWARD_INT:
            yield from self.__regions_contigs[region_ind]
        else:
            for oriented_contig in reversed(self.__regions_contigs[region_ind]):
                yield rev_oriented_contig(oriented_contig)

    def region_path_length(self) -> int:
        """Return the length of all the region paths.

        Returns
        -------
        int
            _description_
        """
        return self.__region_path_length

    # ~*~ Private ~*~

    def __read_regions_contigs(self, contigs_of_regions_path: Path):
        """Extract the oriented contigs for each region.

        Parameters
        ----------
        contigs_of_regions_path : Path
            List of oriented contigs for each region
        """
        vertices = self.__graph.vertices()
        for region_ind, oriented_contigs in enumerate(
                read_contigs_of_regions(contigs_of_regions_path)):
            assert region_ind == vertices.add()
            self.__regions_contigs.append(oriented_contigs)

    def __add_regions_links(self, map_of_regions_path: Path):
        """Add regions links from map of regions.

        Parameters
        ----------
        map_of_regions_path : Path
            Map of regions
        """
        edges = self.__graph.edges()
        reg_indor_iter = read_map_of_regions(map_of_regions_path)
        start_indor = next(reg_indor_iter)
        u_indor = start_indor
        for v_indor in reg_indor_iter:
            edges.add(u_indor, v_indor)
            u_indor = v_indor
        edges.add(u_indor, start_indor)
        self.__region_path_length = len(self.__graph.edges()) // 2


# ============================================================================ #
#                                   FUNCTIONS                                  #
# ============================================================================ #
def rev_oriented_contig(oriented_contig: OrCT) -> OrCT:
    """Return the reverse of the oriented contig.

    Parameters
    ----------
    oriented_contig : OrCT
        Oriented contig

    Returns
    -------
    OrCT
        Its reverse
    """
    return oriented_contig[ORC_ID_IND], ORIENT_REV[oriented_contig[ORC_OR_IND]]


# ---------------------------------------------------------------------------- #
#                               Region Paths File                              #
# ---------------------------------------------------------------------------- #
def write_region_paths(region_paths: Iterable[Iterable[IndOrT]],
                       region_paths_file: Path):
    """Write the region paths file.

    Parameters
    ----------
    region_paths : iterable of iterable of IndOrT
        Region paths
    region_paths_file : Path
        Region paths file path
    """
    with open(region_paths_file, 'w', encoding='utf-8') as f_out:
        for region_path in region_paths:
            line: str = ''
            for reg_ind, reg_or in region_path:
                line += f'{reg_ind}\t{ORIENT_INT_STR[reg_or]}\t'
            f_out.write(line[:-1] + '\n')


def read_region_paths(region_paths_file: Path) -> Iterator[tuple[IndOrT, ...]]:
    """Read region paths from file.

    Parameters
    ----------
    region_paths_file : Path
        Region paths file path

    Yields
    ------
    tuple of IndOrT
        Region path
    """
    with open(region_paths_file, 'r', encoding='utf-8') as regpaths_in:
        for line in regpaths_in:
            region_path = []
            l_regindor = line.split()
            k = 0
            while k < len(l_regindor) - 1:
                region_path.append(
                    (
                        IndexT(l_regindor[k]),
                        STR_ORIENT[l_regindor[k + 1]],  # type: ignore
                    ),
                )
                k += 2
            yield tuple(region_path)


def fmt_region_paths_filename(instance_name: str,
                              ilp_combination: Iterable[RegionIDT]) -> str:
    """Format the region paths filename.

    Parameters
    ----------
    instance_name : str
        Instance name
    ilp_combination : iterable of RegionIDT
        ILP string code combination

    Returns
    -------
    str
        Formatted filename
    """
    return (
        f'{REGION_PATHS_PREFIX}_{instance_name}'
        f"_{'_'.join(ilp_combination)}"
        f'.{REGION_PATHS_EXT}'
    )


# ---------------------------------------------------------------------------- #
#                          Oriented Contig Paths File                          #
# ---------------------------------------------------------------------------- #
def write_oriented_contig_paths(oriented_contig_paths: Iterable[Iterable[OrCT]],
                                oriented_contig_paths_file: Path):
    """Write the oriented contig paths file.

    Parameters
    ----------
    oriented_contig_paths : iterable of iterable of OrCT
        Region paths
    oriented_contig_paths_file : Path
        Region paths file path
    """
    with open(oriented_contig_paths_file, 'w', encoding='utf-8') as f_out:
        for oriented_contig_path in oriented_contig_paths:
            line: str = ''
            for c_id, c_or in oriented_contig_path:
                line += f'{c_id}\t{ORIENT_INT_STR[c_or]}\t'
            f_out.write(line[:-1] + '\n')


def read_oriented_contig_paths(oriented_contig_paths_file: Path) -> (
        Iterator[tuple[OrCT, ...]]):
    """Read region paths from file.

    Parameters
    ----------
    oriented_contig_paths_file : Path
        Region paths file path

    Yields
    ------
    tuple of OrCT
        Oriented contig path
    """
    with open(oriented_contig_paths_file, 'r', encoding='utf-8') as regpaths_in:
        for line in regpaths_in:
            oriented_contig_path = []
            l_orc = line.split()
            k = 0
            while k < len(l_orc) - 1:
                oriented_contig_path.append(
                    (
                        l_orc[k],
                        STR_ORIENT[l_orc[k + 1]],  # type: ignore
                    ),
                )
                k += 2
            yield tuple(oriented_contig_path)


def fmt_oriented_contig_paths_filename(
        instance_name: str,
        ilp_combination: Iterable[RegionIDT]) -> str:
    """Format the oriented contig paths filename.

    Parameters
    ----------
    instance_name : str
        Instance name
    ilp_combination : iterable of RegionIDT
        ILP string code combination

    Returns
    -------
    str
        Formatted filename
    """
    return (
        f'{ORIENTED_CONTIG_PATHS_PREFIX}_{instance_name}'
        f"_{'_'.join(ilp_combination)}"
        f'.{ORIENTED_CONTIG_PATHS_EXT}'
    )



# -*- coding=utf-8 -*-

"""Unit testing for run metadata."""

# pylint: disable=missing-yield-doc,redefined-outer-name,
# flake8: noqa D103

from collections.abc import Iterator
from pathlib import Path

import pytest
import yaml  # type: ignore

from khloraascaf.exceptions import NoSolution
from khloraascaf.inputs import (
    INSTANCE_NAME_DEF,
    MULT_UPB_DEF,
    PRESSCORE_UPB_DEF,
    SOLVER_CBC,
)
from khloraascaf.lib import IR_REGION_ID, SC_REGION_ID
from khloraascaf.run_metadata import (
    IOConfig,
    MetadataAllDebugs,
    MetadataAllSolutions,
    MetadataDebug,
    MetadataSolution,
    str_to_vertex,
    vertex_to_str,
)


# ============================================================================ #
#                                   CONSTANTS                                  #
# ============================================================================ #
TEST_DIR: Path = Path(__file__).parent.absolute()

_TOY_DATADIR = TEST_DIR / 'data'

# ---------------------------------------------------------------------------- #
#                                    IR - SC                                   #
# ---------------------------------------------------------------------------- #
# REFACTOR logic behin tests: do class
_IR_SC_DIR = _TOY_DATADIR / 'ir_sc'
_IR_SC_CONTIG_ATTRS = _IR_SC_DIR / 'contig_attrs.tsv'
_IR_SC_CONTIG_LINKS = _IR_SC_DIR / 'contig_links.tsv'
_IR_SC_CONTIG_STARTER = 'C0'

#
# YAML files
#
_IR_SC_IOCONFIG_YAML = _IR_SC_DIR / 'io_config.yaml'
_IR_SC_SOLUTION_YAML = _IR_SC_DIR / 'solution.yaml'
_IR_SC_SOLUTIONS_YAML = _IR_SC_DIR / 'solutions.yaml'
_IR_SC_DEBUG_YAML = _IR_SC_DIR / 'debug.yaml'
_IR_SC_DEBUGS_YAML = _IR_SC_DIR / 'debugs.yaml'

# REFACTOR uniformise the constants with toy example
_IR_SC_MAP = _IR_SC_DIR / 'map_of_regions_khloraascaf_ir_sc.tsv'
_IR_SC_CTGREG = _IR_SC_DIR / 'contigs_of_regions_khloraascaf_ir_sc.tsv'
_IR_SC_VERTICESREG = _IR_SC_DIR / 'vertices_of_regions_khloraascaf_ir.tsv'
_IR_SC_MAP_DEBUG = _IR_SC_DIR / 'map_of_regions_khloraascaf_ir.tsv'
_IR_SC_REPFRAG = _IR_SC_DIR / 'repfrag_khloraascaf_ir.tsv'


# ============================================================================ #
#                              IOCONFIG CLASS TEST                             #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                    Fixture                                   #
# ---------------------------------------------------------------------------- #
@pytest.fixture
def io_cfg_ir_sc() -> Iterator[IOConfig]:
    """Fixture IOConfig for ir-sc."""
    yield IOConfig.from_run_directory(_IR_SC_DIR)


# ---------------------------------------------------------------------------- #
#                                 Test Methods                                 #
# ---------------------------------------------------------------------------- #
def test_iocfg_from_run_dir(io_cfg_ir_sc: IOConfig):
    assert io_cfg_ir_sc


def test_iocfg_config_directory(io_cfg_ir_sc: IOConfig):
    assert io_cfg_ir_sc.io_config_directory() == _IR_SC_DIR


def test_iocfg_contig_attrs(io_cfg_ir_sc: IOConfig):
    assert (
        io_cfg_ir_sc.contig_attrs()
        == _IR_SC_CONTIG_ATTRS.relative_to(_IR_SC_DIR)
    )


def test_iocfg_contig_links(io_cfg_ir_sc: IOConfig):
    assert (
        io_cfg_ir_sc.contig_links()
        == _IR_SC_CONTIG_LINKS.relative_to(_IR_SC_DIR)
    )


def test_iocfg_contig_starter(io_cfg_ir_sc: IOConfig):
    assert io_cfg_ir_sc.contig_starter() == _IR_SC_CONTIG_STARTER


def test_iocfg_mult_upb(io_cfg_ir_sc: IOConfig):
    assert io_cfg_ir_sc.mult_upb() == MULT_UPB_DEF


def test_iocfg_presscore_upb(io_cfg_ir_sc: IOConfig):
    assert io_cfg_ir_sc.presscore_upb() == PRESSCORE_UPB_DEF


def test_iocfg_solver(io_cfg_ir_sc: IOConfig):
    assert io_cfg_ir_sc.solver() == SOLVER_CBC


def test_iocfg_outdir(io_cfg_ir_sc: IOConfig):
    assert (
        io_cfg_ir_sc.outdir()
        == (_IR_SC_DIR / 'tmp').relative_to(_IR_SC_DIR)
    )


def test_iocfg_instance_name(io_cfg_ir_sc: IOConfig):
    assert io_cfg_ir_sc.instance_name() == INSTANCE_NAME_DEF


def test_iocfg_debug(io_cfg_ir_sc: IOConfig):
    assert io_cfg_ir_sc.debug()


def test_iocfg_write_yaml(io_cfg_ir_sc: IOConfig):
    true_io_cfg_path = _IR_SC_DIR / 'true_io_config.yaml'
    _IR_SC_IOCONFIG_YAML.rename(true_io_cfg_path)
    io_cfg_ir_sc.write_yaml()
    assert _IR_SC_IOCONFIG_YAML.exists()
    with (open(true_io_cfg_path, 'r', encoding='utf-8') as t_in,
            open(_IR_SC_IOCONFIG_YAML, 'r', encoding='utf-8') as f_in):
        assert yaml.load(t_in, yaml.Loader) == yaml.load(f_in, yaml.Loader)
    _IR_SC_IOCONFIG_YAML.unlink()
    true_io_cfg_path.rename(_IR_SC_IOCONFIG_YAML)


def test_iocfg_yaml_path(io_cfg_ir_sc: IOConfig):
    assert io_cfg_ir_sc.yaml_path() == _IR_SC_IOCONFIG_YAML


# ============================================================================ #
#                            METADATASOLUTION CLASS                            #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                    Fixture                                   #
# ---------------------------------------------------------------------------- #
@pytest.fixture
def sol_meta_ir_sc() -> Iterator[MetadataSolution]:
    """Fixture MetadataSolution for ir-sc."""
    yield MetadataSolution.from_run_directory(_IR_SC_DIR)


# ---------------------------------------------------------------------------- #
#                                 Test Methods                                 #
# ---------------------------------------------------------------------------- #
def test_sol_meta_there_is_a_sol():
    assert MetadataSolution.there_is_a_solution(_IR_SC_DIR)


def test_sol_meta_from_run_dir(sol_meta_ir_sc: MetadataSolution):
    assert sol_meta_ir_sc


def test_sol_meta_from_run_dir_fail():
    try:
        assert MetadataSolution.from_run_directory(Path('. /'))
    except NoSolution:
        assert True
    else:
        raise AssertionError


def test_sol_meta_solution_directory(sol_meta_ir_sc: MetadataSolution):
    assert sol_meta_ir_sc.solution_directory() == _IR_SC_DIR


def test_sol_meta_ilp_combination(sol_meta_ir_sc: MetadataSolution):
    assert sol_meta_ir_sc.ilp_combination() == [IR_REGION_ID, SC_REGION_ID]


def test_sol_meta_contigs_of_regions(sol_meta_ir_sc: MetadataSolution):
    assert sol_meta_ir_sc.contigs_of_regions() == Path(_IR_SC_CTGREG.name)


def test_sol_meta_map_of_regions(sol_meta_ir_sc: MetadataSolution):
    assert sol_meta_ir_sc.map_of_regions() == Path(_IR_SC_MAP.name)


def test_sol_meta_write_yaml(sol_meta_ir_sc: MetadataSolution):
    true_sol_meta_path = _IR_SC_DIR / 'true_solution.yaml'
    _IR_SC_SOLUTION_YAML.rename(true_sol_meta_path)
    sol_meta_ir_sc.write_yaml()
    assert _IR_SC_SOLUTION_YAML.exists()
    with (open(true_sol_meta_path, 'r', encoding='utf-8') as t_in,
            open(_IR_SC_SOLUTION_YAML, 'r', encoding='utf-8') as f_in):
        assert yaml.load(t_in, yaml.Loader) == yaml.load(f_in, yaml.Loader)
    _IR_SC_SOLUTION_YAML.unlink()
    true_sol_meta_path.rename(_IR_SC_SOLUTION_YAML)


def test_sol_meta_yaml_path(sol_meta_ir_sc: MetadataSolution):
    assert sol_meta_ir_sc.yaml_path() == _IR_SC_SOLUTION_YAML


# ============================================================================ #
#                          METADATAALLSOLUTIONS CLASS                          #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                    Fixture                                   #
# ---------------------------------------------------------------------------- #
@pytest.fixture
def allsol_meta_ir_sc() -> Iterator[MetadataAllSolutions]:
    """Fixture MetadataAllSolutions for ir-sc."""
    yield MetadataAllSolutions.from_run_directory(_IR_SC_DIR)


# ---------------------------------------------------------------------------- #
#                                 Test Methods                                 #
# ---------------------------------------------------------------------------- #
def test_allsol_meta_there_is_a_sol():
    assert MetadataAllSolutions.there_is_a_solution(_IR_SC_DIR)


def test_allsol_meta_from_run_dir(allsol_meta_ir_sc: MetadataAllSolutions):
    assert allsol_meta_ir_sc


def test_sol_meta_from_run_dir_empty():
    nosol_meta = MetadataAllSolutions.from_run_directory(Path('. /'))
    assert not nosol_meta


def test_allsol_meta_solutions_directory(
        allsol_meta_ir_sc: MetadataAllSolutions):
    assert allsol_meta_ir_sc.solutions_directory() == _IR_SC_DIR


def test_allsol_meta_iter(allsol_meta_ir_sc: MetadataAllSolutions):
    assert sum(1 for _ in allsol_meta_ir_sc) == 1


def test_allsol_meta_len(allsol_meta_ir_sc: MetadataAllSolutions):
    assert len(allsol_meta_ir_sc) == 1


def test_allsol_meta_write_yaml(allsol_meta_ir_sc: MetadataAllSolutions):
    true_allsol_meta_path = _IR_SC_DIR / 'true_solutions.yaml'
    _IR_SC_SOLUTIONS_YAML.rename(true_allsol_meta_path)
    allsol_meta_ir_sc.write_yaml()
    assert _IR_SC_SOLUTIONS_YAML.exists()
    with (open(true_allsol_meta_path, 'r', encoding='utf-8') as t_in,
            open(_IR_SC_SOLUTIONS_YAML, 'r', encoding='utf-8') as f_in):
        assert yaml.load(t_in, yaml.Loader) == yaml.load(f_in, yaml.Loader)
    _IR_SC_SOLUTIONS_YAML.unlink()
    true_allsol_meta_path.rename(_IR_SC_SOLUTIONS_YAML)


def test_allsol_meta_yaml_path(allsol_meta_ir_sc: MetadataAllSolutions):
    assert allsol_meta_ir_sc.yaml_path() == _IR_SC_SOLUTIONS_YAML


# ============================================================================ #
#                              METADATADEBUG CLASS                             #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                    Fixture                                   #
# ---------------------------------------------------------------------------- #
@pytest.fixture
def debug_meta_ir_sc() -> Iterator[MetadataDebug]:
    """Fixture MetadataDebug for ir-sc."""
    yield MetadataDebug.from_run_directory(_IR_SC_DIR)


# ---------------------------------------------------------------------------- #
#                                 Test Methods                                 #
# ---------------------------------------------------------------------------- #
def test_debug_meta_from_run_dir(debug_meta_ir_sc: MetadataDebug):
    assert debug_meta_ir_sc


def test_debug_meta_run_directory(debug_meta_ir_sc: MetadataDebug):
    assert debug_meta_ir_sc.run_directory() == _IR_SC_DIR


def test_debug_meta_ilp_combination(debug_meta_ir_sc: MetadataDebug):
    assert debug_meta_ir_sc.ilp_combination() == [IR_REGION_ID]


def test_debug_meta_starter_vertex(debug_meta_ir_sc: MetadataDebug):
    assert debug_meta_ir_sc.starter_vertex() == (0, 0, 0)


def test_debug_meta_ilp_status(debug_meta_ir_sc: MetadataDebug):
    assert debug_meta_ir_sc.ilp_status() == 'Optimal'


def test_debug_meta_opt_value(debug_meta_ir_sc: MetadataDebug):
    assert debug_meta_ir_sc.opt_value() == 3.0


def test_debug_meta_vertices_of_regions(debug_meta_ir_sc: MetadataDebug):
    assert debug_meta_ir_sc.vertices_of_regions() == Path(
        _IR_SC_VERTICESREG.name)


def test_debug_meta_map_of_regions(debug_meta_ir_sc: MetadataDebug):
    assert debug_meta_ir_sc.map_of_regions() == Path(_IR_SC_MAP_DEBUG.name)


def test_debug_meta_repeat_fragments(debug_meta_ir_sc: MetadataDebug):
    assert debug_meta_ir_sc.repeat_fragments() == Path(_IR_SC_REPFRAG.name)


def test_debug_meta_write_yaml(debug_meta_ir_sc: MetadataDebug):
    true_debug_meta_path = _IR_SC_DIR / 'true_debug.yaml'
    _IR_SC_DEBUG_YAML.rename(true_debug_meta_path)
    debug_meta_ir_sc.write_yaml()
    assert _IR_SC_DEBUG_YAML.exists()
    with (open(true_debug_meta_path, 'r', encoding='utf-8') as t_in,
            open(_IR_SC_DEBUG_YAML, 'r', encoding='utf-8') as f_in):
        assert yaml.load(t_in, yaml.Loader) == yaml.load(f_in, yaml.Loader)
    _IR_SC_DEBUG_YAML.unlink()
    true_debug_meta_path.rename(_IR_SC_DEBUG_YAML)


def test_debug_meta_yaml_path(debug_meta_ir_sc: MetadataDebug):
    assert debug_meta_ir_sc.yaml_path() == _IR_SC_DEBUG_YAML


# ============================================================================ #
#                               METADATAALLDEBUGS                              #
# ============================================================================ #
# ---------------------------------------------------------------------------- #
#                                    Fixture                                   #
# ---------------------------------------------------------------------------- #
@pytest.fixture
def alldebugs_meta_ir_sc() -> Iterator[MetadataAllDebugs]:
    """Fixture MetadataAllDebugs for ir-sc."""
    yield MetadataAllDebugs.from_run_directory(_IR_SC_DIR)


# ---------------------------------------------------------------------------- #
#                                 Test Methods                                 #
# ---------------------------------------------------------------------------- #
def test_alldebugs_meta_from_run_dir(alldebugs_meta_ir_sc: MetadataAllDebugs):
    assert alldebugs_meta_ir_sc


def test_alldebugs_meta_run_dir(alldebugs_meta_ir_sc: MetadataAllDebugs):
    assert alldebugs_meta_ir_sc.runs_directory() == _IR_SC_DIR


def test_alldebugs_meta_iter(alldebugs_meta_ir_sc: MetadataAllDebugs):
    assert sum(1 for _ in alldebugs_meta_ir_sc) == 4


def test_alldebugs_meta_len(alldebugs_meta_ir_sc: MetadataAllDebugs):
    assert len(alldebugs_meta_ir_sc) == 4


def test_alldebugs_meta_write_yaml(alldebugs_meta_ir_sc: MetadataAllDebugs):
    true_alldebugs_meta_path = _IR_SC_DIR / 'true_debugs.yaml'
    _IR_SC_DEBUGS_YAML.rename(true_alldebugs_meta_path)
    alldebugs_meta_ir_sc.write_yaml()
    assert _IR_SC_DEBUGS_YAML.exists()
    with (open(true_alldebugs_meta_path, 'r', encoding='utf-8') as t_in,
            open(_IR_SC_DEBUGS_YAML, 'r', encoding='utf-8') as f_in):
        assert yaml.load(t_in, yaml.Loader) == yaml.load(f_in, yaml.Loader)
    _IR_SC_DEBUGS_YAML.unlink()
    true_alldebugs_meta_path.rename(_IR_SC_DEBUGS_YAML)


def test_alldebugs_meta_yaml_path(alldebugs_meta_ir_sc: MetadataAllDebugs):
    assert alldebugs_meta_ir_sc.yaml_path() == _IR_SC_DEBUGS_YAML


# ============================================================================ #
#                                FUNCTIONS TESTS                               #
# ============================================================================ #
def test_vertex_to_str():
    assert vertex_to_str((1, 0, 4)) == '1\t+\t4'
    assert vertex_to_str((8, 1, 0)) == '8\t-\t0'


def test_str_to_vertex():
    assert str_to_vertex('1\t+\t4') == (1, 0, 4)
    assert str_to_vertex('8\t-\t0') == (8, 1, 0)
